# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack import nb_exceptions
from ovn_bgp_agent.drivers.openstack.watchers import nb_bgp_watcher
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.tests import utils


class TestLogicalSwitchPortProviderCreateEvent(test_base.TestCase):

    def setUp(self):
        super(TestLogicalSwitchPortProviderCreateEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = {
            'net1': ['10.0.0.5']}

        # Assume the logical switch has been setup properly.
        self.agent.is_ls_provider.return_value = True

        # Assume the ip is not exposed yet
        self.agent.is_ip_exposed.return_value = False

        self.event = nb_bgp_watcher.LogicalSwitchPortProviderCreateEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        old = utils.create_row(options={}, up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_port_up(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        old = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_external_id(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               external_ids={'neutron:host_id': self.chassis},
                               up=[True])
        old = utils.create_row(external_ids={}, up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_invalid_address(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac '],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_chassis(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': 'fake_chassis'},
                               up=[True])
        old = utils.create_row(options={}, up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_tenant_create(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        old = utils.create_row(options={}, up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_additional_bindings(self):
        event = self.event.ROW_UPDATE
        bindings = ','.join([self.chassis, 'other-chassis'])
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': bindings},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': self.chassis})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            up=[True])
        self.assertIsNone(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            external_ids={'neutron:host_id': self.chassis,
                          constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'},
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            up=[True])
        ips_info = {
            'mac': 'mac',
            'cidrs': [],
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_switch': 'test-ls',
        }
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_ip.assert_called_once_with(['192.168.0.1'], ips_info)


class TestLogicalSwitchPortProviderDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestLogicalSwitchPortProviderDeleteEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.agent = mock.Mock(chassis=self.chassis)
        self.agent.ovn_local_lrps = {
            'net1': ['10.0.0.5']}

        # Assume the logical switch has been setup properly.
        self.agent.is_ls_provider.return_value = True

        # Assume the ip is exposed
        self.agent.is_ip_exposed.return_value = True

        self.event = nb_bgp_watcher.LogicalSwitchPortProviderDeleteEvent(
            self.agent)

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[False])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_chassis(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               external_ids={'neutron:host_id': 'chassis2'},
                               up=[True])
        old = utils.create_row(external_ids={'neutron:host_id': self.chassis},
                               up=[True])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_status_different_chassis(self):
        # Update test assumption, since the ip should not be exposed
        self.agent.is_ip_exposed.return_value = False

        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': 'other-chassis'},
                               up=[False])
        old = utils.create_row(options={'requested-chassis': 'other-chassis'},
                               up=[True])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[False])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_ignore_not_up_with_additional_bindings(self):
        event = self.event.ROW_UPDATE
        bindings = ','.join([self.chassis, 'other-chassis'])
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': bindings},
                               up=[False])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_invalid_address(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac '],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_chassis(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': 'other_chassis'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_tenant_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        self.assertFalse(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            up=[True])
        self.assertIsNone(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            external_ids={'neutron:host_id': self.chassis,
                          constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'},
            addresses=['mac 192.168.0.1'],
            options={'requested-chassis': self.chassis},
            up=[True])
        ips_info = {
            'mac': 'mac',
            'cidrs': [],
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_switch': 'test-ls'
        }

        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.withdraw_ip.assert_called_once_with(['192.168.0.1'],
                                                       ips_info)


class TestLogicalSwitchPortFIPCreateEvent(test_base.TestCase):

    def setUp(self):
        super(TestLogicalSwitchPortFIPCreateEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.agent = mock.Mock(chassis=self.chassis)

        # Assume the logical switch has been setup properly.
        self.agent.is_ls_provider.return_value = True

        # Assume the ip is not exposed yet
        self.agent.is_ip_exposed.return_value = False

        self.event = nb_bgp_watcher.LogicalSwitchPortFIPCreateEvent(
            self.agent)

    def test_match_fn_chassis_change(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        old = utils.create_row(options={}, up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_chassis_change_external_ids(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            external_ids={
                constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis,
                constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
            up=[True])
        old = utils.create_row(external_ids={
            constants.OVN_FIP_EXT_ID_KEY: 'fip-ip',
        }, up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_change_external_ids(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            external_ids={
                constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis,
                constants.OVN_FIP_EXT_ID_KEY: 'fip-ip',
                'neutron:revision_number': '416',
            },
            up=[True])
        old = utils.create_row(
            external_ids={
                constants.OVN_FIP_EXT_ID_KEY: 'fip-ip',
                constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis,
                'neutron:revision_number': '417',
            },
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_status_change(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_fip_addition(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               external_ids={},
                               up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_fip(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_chassis(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': 'wrong_chassis'},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_port_down(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_address(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac '],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_exception(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        self.assertIsNone(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        external_ip = '10.0.0.10'
        ls_name = 'neutron-net-id'
        self.agent.get_port_external_ip_and_ls.return_value = (external_ip,
                                                               'mac',
                                                               ls_name)
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               name='net-id')
        self.event.run(mock.Mock(), row, mock.Mock())
        self.agent.expose_fip.assert_called_once_with(external_ip, 'mac',
                                                      ls_name, row)

    def test_run_no_external_ip(self):
        self.agent.get_port_external_ip_and_ls.side_effect = (
            nb_exceptions.NATNotFound)

        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               name='net-id')
        self.event.run(mock.ANY, row, mock.ANY)
        self.agent.expose_fip.assert_not_called()


class TestLogicalSwitchPortFIPDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestLogicalSwitchPortFIPDeleteEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.agent = mock.Mock(chassis=self.chassis)

        # Assume the logical switch has been setup properly.
        self.agent.is_ls_provider.return_value = True

        # Assume the ip is exposed
        self.agent.is_ip_exposed.return_value = True

        self.event = nb_bgp_watcher.LogicalSwitchPortFIPDeleteEvent(
            self.agent)

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        self.assertTrue(self.event.match_fn(event, row, utils.create_row()))

    def test_match_fn_update(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[False])
        old = utils.create_row(up=[True])
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_different_chassis(self):
        # Update test assumption, since the ip should not be exposed
        self.agent.is_ip_exposed.return_value = False

        event = self.event.ROW_UPDATE
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': 'other-chassis'},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[False])
        old = utils.create_row(up=[True])
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_update_external_id(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            external_ids={
                constants.OVN_HOST_ID_EXT_ID_KEY: 'other-chassis',
                constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
            up=[True])
        old = utils.create_row(external_ids={
            constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis,
            constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_external_id_remove_fip(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            external_ids={
                constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis},
            up=[True])
        old = utils.create_row(external_ids={
            constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis,
            constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'})
        self.assertTrue(self.event.match_fn(event, row, old))

    def test_match_fn_update_external_id_no_fip(self):
        event = self.event.ROW_UPDATE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 192.168.0.1'],
            external_ids={
                constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis},
            up=[True])
        old = utils.create_row(external_ids={
            constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis})
        self.assertFalse(self.event.match_fn(event, row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               up=[False])
        old = utils.create_row()
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_up(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[False])
        old = utils.create_row(options={'requested-chassis': self.chassis},
                               up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_invalid_address(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac '],
                               options={'requested-chassis': self.chassis},
                               up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_chassis(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': 'other_chassis'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_chassis_update(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': 'other_chassis'},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        old = utils.create_row(options={'requested-chassis': self.chassis})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_fip_update(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               addresses=['mac 192.168.0.1'],
                               options={'requested-chassis': self.chassis},
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'new-fip-ip'},
                               up=[True])
        old = utils.create_row(
            external_ids={constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE)
        self.assertIsNone(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               external_ids={
                                   constants.OVN_FIP_EXT_ID_KEY: 'fip-ip'},
                               up=[True])
        self.event.run(mock.Mock(), row, utils.create_row())
        self.agent.withdraw_fip.assert_called_once_with('fip-ip', row)

    def test_run_no_fip(self):
        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               external_ids={})
        old = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE,
                               external_ids={})
        self.event.run(mock.Mock(), row, old)
        self.agent.withdraw_fip.assert_not_called()


class TestLogicalSwitchUpdateEvent(test_base.TestCase):

    def setUp(self):
        super(TestLogicalSwitchUpdateEvent, self).setUp()
        self.agent = mock.Mock()
        self.event = nb_bgp_watcher.LogicalSwitchUpdateEvent(
            self.agent)
        self.row = utils.create_row(external_ids={
            constants.OVN_EVPN_TYPE_EXT_ID_KEY: constants.OVN_EVPN_TYPE_L3,
            constants.OVN_EVPN_VNI_EXT_ID_KEY: 1001,
        })

    def test_match_fn(self):
        # ls had no conf in external_ids, does have it now.
        old = utils.create_row(external_ids={})
        self.assertTrue(self.event.match_fn(None, self.row, old))

    def test_match_fn_changed_vni(self):
        old = utils.create_row(external_ids={
            constants.OVN_EVPN_TYPE_EXT_ID_KEY: constants.OVN_EVPN_TYPE_L3,
            constants.OVN_EVPN_VNI_EXT_ID_KEY: 1002,
        })
        self.assertTrue(self.event.match_fn(None, self.row, old))

    def test_match_fn_deleted_ls(self):
        self.assertTrue(self.event.match_fn(self.event.ROW_DELETE, self.row,
                                            None))

    def test_match_fn_no_match(self):
        # no vrf update in old, should return False
        old = utils.create_row()
        self.assertFalse(self.event.match_fn(None, self.row, old))

    def test_match_fn_no_match_same_vni(self):
        # same update in old, should return False
        self.assertFalse(self.event.match_fn(None, self.row, self.row))

    def test_match_fn_no_match_incomplete_row(self):
        # incomplete configuration, should return False
        row = utils.create_row(external_ids={
            constants.OVN_EVPN_TYPE_EXT_ID_KEY: constants.OVN_EVPN_TYPE_L3,
        })
        self.assertFalse(self.event.match_fn(None, row, None))

        row = utils.create_row(external_ids={
            constants.OVN_EVPN_VNI_EXT_ID_KEY: 1001,
        })
        self.assertFalse(self.event.match_fn(None, row, None))

    def test_run(self):
        self.event.run(None, None, None)
        self.agent.sync.assert_called_once()


class TestLocalnetCreateDeleteEvent(test_base.TestCase):

    def setUp(self):
        super(TestLocalnetCreateDeleteEvent, self).setUp()
        self.agent = mock.Mock()
        self.event = nb_bgp_watcher.LocalnetCreateDeleteEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(type=constants.OVN_LOCALNET_VIF_PORT_TYPE)
        self.assertTrue(self.event.match_fn(None, row, None))

        row = utils.create_row(type=constants.OVN_VM_VIF_PORT_TYPE)
        self.assertFalse(self.event.match_fn(None, row, None))

    def test_run(self):
        row = utils.create_row(type=constants.OVN_LOCALNET_VIF_PORT_TYPE)
        self.event.run(None, row, None)
        self.agent.sync.assert_called_once()


class TestChassisRedirectCreateEvent(test_base.TestCase):
    def setUp(self):
        super(TestChassisRedirectCreateEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.event = nb_bgp_watcher.ChassisRedirectCreateEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        old = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_no_status_change(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        old = utils.create_row()
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_different_chassis(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': 'other_chassis'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_no_networks(self):
        row = utils.create_row(
            mac='fake-mac',
            networks=[],
            status={'hosting-chassis': self.chassis_id},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'})
        self.assertFalse(self.event.match_fn(None, row, None))

    def test_run(self):
        row = utils.create_row(
            mac='fake-mac',
            networks=['192.168.0.2/24'],
            status={'hosting-chassis': self.chassis_id},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'})
        ips_info = {'mac': 'fake-mac',
                    'cidrs': ['192.168.0.2/24'],
                    'type': constants.OVN_CR_LRP_PORT_TYPE,
                    'logical_switch': 'test-ls',
                    'router': None}
        self.event.run(None, row, None)
        self.agent.expose_ip.assert_called_once_with(['192.168.0.2'], ips_info)


class TestChassisRedirectDeleteEvent(test_base.TestCase):
    def setUp(self):
        super(TestChassisRedirectDeleteEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.event = nb_bgp_watcher.ChassisRedirectDeleteEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={})
        old = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_exception(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_no_status_change(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        old = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_different_chassis(self):
        row = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': self.chassis_id})
        old = utils.create_row(mac='fake-mac',
                               networks=['192.168.0.2/24'],
                               status={'hosting-chassis': 'different_chassis'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_networks(self):
        row = utils.create_row(
            mac='fake-mac',
            networks=[],
            status={'hosting-chassis': self.chassis_id},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'})
        self.assertFalse(self.event.match_fn(None, row, None))

    def test_run(self):
        row = utils.create_row(
            mac='fake-mac',
            networks=['192.168.0.2/24'],
            status={'hosting-chassis': self.chassis_id},
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'test-ls'})
        ips_info = {'mac': 'fake-mac',
                    'cidrs': ['192.168.0.2/24'],
                    'type': constants.OVN_CR_LRP_PORT_TYPE,
                    'logical_switch': 'test-ls',
                    'router': None}
        self.event.run(None, row, None)
        self.agent.withdraw_ip.assert_called_once_with(['192.168.0.2'],
                                                       ips_info)


class TestLogicalSwitchPortSubnetAttachEvent(test_base.TestCase):
    def setUp(self):
        super(TestLogicalSwitchPortSubnetAttachEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_cr_lrps = {
            'router1': {'bridge_device': 'br-ex',
                        'bridge_vlan': None,
                        'ips': ['172.24.16.2']}}
        self.event = nb_bgp_watcher.LogicalSwitchPortSubnetAttachEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        old = utils.create_row(up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_associate_router(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        old = utils.create_row(
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_device_owner(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_gateway',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_local_crlrp(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router2'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.24.1/24",
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'network1',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        subnet_info = {
            'associated_router': 'router1',
            'network': 'network1',
            'address_scopes': {4: None, 6: None}}
        self.event.run(None, row, None)
        self.agent.expose_subnet.assert_called_once_with(["192.168.24.1/24"],
                                                         subnet_info)


class TestLogicalSwitchPortSubnetDetachEvent(test_base.TestCase):
    def setUp(self):
        super(TestLogicalSwitchPortSubnetDetachEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_cr_lrps = {
            'router1': {'bridge_device': 'br-ex',
                        'bridge_vlan': None,
                        'ips': ['172.24.16.2']}}
        self.event = nb_bgp_watcher.LogicalSwitchPortSubnetDetachEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[False])
        old = utils.create_row(up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_delete_down(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[False])
        self.assertFalse(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_disassociate_router(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface'},
            up=[True])
        old = utils.create_row(
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_exception(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_wrong_device_owner(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_gateway',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        old = utils.create_row(up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_local_crlrp(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface'},
            up=[True])
        old = utils.create_row(
            external_ids={
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                    constants.OVN_DEVICE_ID_EXT_ID_KEY: 'other_router'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.24.1/24",
                constants.OVN_LS_NAME_EXT_ID_KEY: 'network1',
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface'},
            up=[True])
        old = utils.create_row(
            external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.24.1/24",
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'network1',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'})
        subnet_info = {
            'associated_router': 'router1',
            'network': 'network1',
            'address_scopes': {4: None, 6: None}}
        self.event.run(None, row, old)
        self.agent.withdraw_subnet.assert_called_once_with(
            ["192.168.24.1/24"], subnet_info)

    def test_run_no_old_external_ids(self):
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.24.1/24",
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'network1',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        old = utils.create_row()
        subnet_info = {
            'associated_router': 'router1',
            'network': 'network1',
            'address_scopes': {4: None, 6: None}}
        self.event.run(None, row, old)
        self.agent.withdraw_subnet.assert_called_once_with(
            ["192.168.24.1/24"], subnet_info)

    def test_run_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_ROUTER_PORT_TYPE,
            external_ids={
                constants.OVN_CIDRS_EXT_ID_KEY: "192.168.24.1/24",
                constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                    'network:router_interface',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'network1',
                constants.OVN_DEVICE_ID_EXT_ID_KEY: 'router1'},
            up=[True])
        subnet_info = {
            'associated_router': 'router1',
            'network': 'network1',
            'address_scopes': {4: None, 6: None}}
        self.event.run(event, row, mock.Mock())
        self.agent.withdraw_subnet.assert_called_once_with(
            ["192.168.24.1/24"], subnet_info)


class TestLogicalSwitchPortTenantCreateEvent(test_base.TestCase):
    def setUp(self):
        super(TestLogicalSwitchPortTenantCreateEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_lrps = {
            'net1': ['10.0.0.5']}
        self.event = nb_bgp_watcher.LogicalSwitchPortTenantCreateEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        old = utils.create_row(up=[False])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_network_set(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        old = utils.create_row(external_ids={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_wong_ip(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_local_lrp(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net2'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_exception(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                          constants.OVN_CIDRS_EXT_ID_KEY: "10.0.0.6/24"},
            up=[True])
        ips_info = {
            'mac': 'mac',
            'cidrs': ["10.0.0.6/24"],
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_switch': 'net1'}
        self.event.run(None, row, mock.Mock())
        self.agent.expose_remote_ip.assert_called_once_with(
            ["10.0.0.6"], ips_info)

    def test_run_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                          constants.OVN_CIDRS_EXT_ID_KEY: "10.0.0.6/24"},
            up=[True])
        self.event.run(None, row, mock.Mock())
        self.agent.expose_remote_ip.assert_not_called()


class TestLogicalSwitchPortTenantDeleteEvent(test_base.TestCase):
    def setUp(self):
        super(TestLogicalSwitchPortTenantDeleteEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_lrps = {
            'net1': ['10.0.0.5']}
        self.event = nb_bgp_watcher.LogicalSwitchPortTenantDeleteEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[False])
        old = utils.create_row(up=[True])
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_wong_ip(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_not_up(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            up=[True])
        old = utils.create_row(up=[False])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_not_local_lrp(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net2'},
            up=[True])
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_exception(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run(self):
        row = utils.create_row(
            type=constants.OVN_VM_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                          constants.OVN_CIDRS_EXT_ID_KEY: "10.0.0.6/24"},
            up=[True])
        ips_info = {
            'mac': 'mac',
            'cidrs': ["10.0.0.6/24"],
            'type': constants.OVN_VM_VIF_PORT_TYPE,
            'logical_switch': 'net1'}
        self.event.run(None, row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_called_once_with(
            ["10.0.0.6"], ips_info)

    def test_run_wrong_type(self):
        row = utils.create_row(
            type=constants.OVN_PATCH_VIF_PORT_TYPE,
            addresses=['mac 10.0.0.6'],
            external_ids={constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                          constants.OVN_CIDRS_EXT_ID_KEY: "10.0.0.6/24"},
            up=[True])
        self.event.run(None, row, mock.Mock())
        self.agent.withdraw_remote_ip.assert_not_called()


class TestOVNLBCreateEvent(test_base.TestCase):
    def setUp(self):
        super(TestOVNLBCreateEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_cr_lrps = {
            'router1': {'bridge_device': 'br-ex',
                        'bridge_vlan': None,
                        'ips': ['172.24.16.2']}}
        self.event = nb_bgp_watcher.OVNLBCreateEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(vips={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_router_added(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(external_ids={})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_fip_added(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1'},
            vips={'192.168.1.50:80': '192.168.1.100:80'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_vips_no_change(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_vips(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_match_fn_no_local_crlrp(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router2',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, mock.Mock()))

    def test_run_vip(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80'})
        old = utils.create_row(vips={})

        self.event.run(None, row, old)

        self.agent.expose_ovn_lb_vip.assert_called_once_with(row)
        self.agent.expose_ovn_lb_fip.assert_not_called()

    def test_run_vip_and_fip(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(vips={})

        self.event.run(None, row, old)

        self.agent.expose_ovn_lb_vip.assert_called_once_with(row)
        self.agent.expose_ovn_lb_fip.assert_called_once_with(row)

    def test_run_vip_added_router(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                'other': 'info'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.event.run(None, row, old)

        self.agent.expose_ovn_lb_vip.assert_called_once_with(row)
        self.agent.expose_ovn_lb_fip.assert_not_called()

    def test_run_fip(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(external_ids={
            constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1'})

        self.event.run(None, row, old)

        self.agent.expose_ovn_lb_vip.assert_not_called()
        self.agent.expose_ovn_lb_fip.assert_called_once_with(row)


class TestOVNLBDeleteEvent(test_base.TestCase):
    def setUp(self):
        super(TestOVNLBDeleteEvent, self).setUp()
        self.chassis = 'fake-chassis'
        self.chassis_id = 'fake-chassis-id'
        self.agent = mock.Mock(chassis=self.chassis,
                               chassis_id=self.chassis_id)
        self.agent.ovn_local_cr_lrps = {
            'router1': {'bridge_device': 'br-ex',
                        'bridge_vlan': None,
                        'ips': ['172.24.16.2']}}
        self.event = nb_bgp_watcher.OVNLBDeleteEvent(
            self.agent)

    def test_match_fn(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={})
        old = utils.create_row(
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertTrue(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_delete_no_vips(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={})
        self.assertFalse(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_delete_no_local_router(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router2',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertFalse(self.event.match_fn(event, row, mock.Mock()))

    def test_match_fn_router_deleted(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(external_ids={
            constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1'
        })
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_no_old_router(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(external_ids={})
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_old_router_non_local(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        old = utils.create_row(external_ids={
            constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router2',
        })
        self.assertFalse(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_fip_deleted(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50'},
            vips={'192.168.1.50:80': '192.168.1.100:80'})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_match_fn_vip_deleted_with_ext_id_update(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        self.assertTrue(self.event.match_fn(mock.Mock(), row, old))

    def test_run_vip_delete_without_external_ids_on_old(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={})
        old = utils.create_row(vips={'192.168.1.50:80': '192.168.1.100:80'})

        self.event.run(None, row, old)

        self.agent.withdraw_ovn_lb_vip.assert_not_called()
        self.agent.withdraw_ovn_lb_fip.assert_not_called()

    def test_run_vip_delete(self):
        event = self.event.ROW_DELETE
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50'},
            vips={'192.168.1.50:80': '192.168.1.100:80'})
        self.event.run(event, row, None)

        self.agent.withdraw_ovn_lb_vip.assert_called_once_with(row)

    def test_run_vip_deleted_extra_ext_id_info(self):
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LB_VIP_IP_EXT_ID_KEY: '192.168.1.50',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1',
                'other': 'info'},
            vips={})

        self.event.run(None, row, old)

        self.agent.withdraw_ovn_lb_vip.assert_called_once_with(old)
        self.agent.withdraw_ovn_lb_fip.assert_called_once_with(old)

    def test_run_fip(self):
        row = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LS_NAME_EXT_ID_KEY: 'net1'},
            vips={'192.168.1.50:80': '192.168.1.100:80'})
        old = utils.create_row(
            external_ids={
                constants.OVN_LB_LR_REF_EXT_ID_KEY: 'neutron-router1',
                constants.OVN_LB_VIP_FIP_EXT_ID_KEY: '172.24.4.5'},
            vips={'192.168.1.50:80': '192.168.1.100:80',
                  '172.24.4.5:80': '192.168.1.100:80'})

        self.event.run(None, row, old)

        self.agent.withdraw_ovn_lb_vip.assert_not_called()
        self.agent.withdraw_ovn_lb_fip.assert_called_once_with(old)


class TestNATMACAddedEvent(test_base.TestCase):
    def setUp(self):
        super().setUp()
        self.chassis = 'fake-chassis'
        self.agent = mock.Mock(chassis=self.chassis)
        self.event = nb_bgp_watcher.NATMACAddedEvent(
            self.agent)
        self.nat_entry = utils.create_row(
            _uuid='uuid',
            logical_port=['port-id'],
            external_ids={constants.OVN_FIP_NET_EXT_ID_KEY: 'net-id'},
            external_ip='ext-ip',
            external_mac=['ext-mac']
        )

    def _call_match(self, old=mock.ANY):
        return self.event.match_fn(mock.ANY, self.nat_entry, old)

    def test_run(self):
        self.event.run(mock.ANY, self.nat_entry, None)

    def test_match_fn_no_logical_port(self):
        self.nat_entry.logical_port = ""
        self.assertFalse(self._call_match())

    def test_match_fn_lsp_not_found(self):
        self.agent.nb_idl.lsp_get.return_value.execute.return_value = None
        self.assertFalse(self._call_match())

    def test_match_fn_bad_lsp_type(self):
        self.agent.nb_idl.lsp_get.return_value.execute.return_value.type = 'b'
        self.assertFalse(self._call_match())

    def test_match_fn_no_requested_chassis(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {}
        self.assertFalse(self._call_match())

    def test_match_fn_different_chassis(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': 'foo'}
        self.assertFalse(self._call_match())

    def test_match_fn_different_chassis_virtual_port(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VIRTUAL_VIF_PORT_TYPE
        lsp.external_ids = {constants.OVN_HOST_ID_EXT_ID_KEY: 'foo'}
        self.assertFalse(self._call_match())

    def test_match_fn_external_mac_untouched(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
        )
        self.assertFalse(self._call_match(old))

    def test_match_fn_external_mac_unset(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=['old-mac']
        )
        self.nat_entry.external_mac = ""
        self.assertFalse(self._call_match(old))

    def test_match_fn_external_mac_already_set(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=self.nat_entry.external_mac
        )
        self.assertFalse(self._call_match(old))

    def test_match_fn_ext_ids_missing_net_id(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=""
        )
        del self.nat_entry.external_ids[constants.OVN_FIP_NET_EXT_ID_KEY]
        self.assertFalse(self._call_match(old))

    def test_match_fn_missing_ext_ip(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=""
        )
        self.nat_entry.external_ip = ""
        self.assertFalse(self._call_match(old))

    def test_match_fn_passed_virtual_port(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VIRTUAL_VIF_PORT_TYPE
        lsp.external_ids = {constants.OVN_HOST_ID_EXT_ID_KEY: self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=""
        )
        self.assertTrue(self._call_match(old))

    def test_match_fn_passed(self):
        lsp = self.agent.nb_idl.lsp_get.return_value.execute.return_value
        lsp.type = constants.OVN_VM_VIF_PORT_TYPE
        lsp.options = {'requested-chassis': self.chassis}
        old = utils.create_row(
            _uuid='uuid',
            external_mac=""
        )
        self.assertTrue(self._call_match(old))



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from ovsdbapp.schema.open_vswitch import impl_idl as idl_ovs

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import ovs as ovs_utils
from ovn_bgp_agent import exceptions as agent_exc
from ovn_bgp_agent.tests import base as test_base
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF


class TestOVS(test_base.TestCase):

    def setUp(self):
        super(TestOVS, self).setUp()
        self.mock_ovs_vsctl = mock.patch(
            'ovn_bgp_agent.privileged.ovs_vsctl').start()

        # Helper variables that are used across multiple methods
        self.bridge = 'br-fake'
        self.flows_info = {self.bridge: {'in_port': set()}}
        self.cookie = 'fake-cookie'
        self.cookie_id = 'cookie=%s/-1' % self.cookie
        self.mac = 'aa:bb:cc:dd:ee:ff'

    def _test_get_bridge_flows(self, has_filter=False):
        port_iface = '1'
        fake_flow_0 = '{},ip,in_port={}'.format(self.cookie_id, port_iface)
        fake_flow_1 = '{},ipv6,in_port={}'.format(self.cookie_id, port_iface)
        fake_filter = 'cookie=fake-cookie/-1' if has_filter else None
        flows = 'HEADER\n%s\n%s\n' % (fake_flow_0, fake_flow_1)
        self.mock_ovs_vsctl.ovs_ofctl.return_value = [flows]

        ret = ovs_utils.get_bridge_flows(self.bridge, filter_=fake_filter)

        expected_args = ['dump-flows', self.bridge]
        if has_filter:
            expected_args.append(fake_filter)

        self.mock_ovs_vsctl.ovs_ofctl.assert_called_once_with(expected_args)
        self.assertEqual([fake_flow_0, fake_flow_1], ret)

    def test_get_bridge_flows(self):
        self._test_get_bridge_flows()

    def test_get_bridge_flows_with_filters(self):
        self._test_get_bridge_flows(has_filter=True)

    def test_get_device_port_at_ovs(self):
        port = 'fake-port'
        port_iface = '1'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = port_iface

        ret = ovs_utils.get_device_port_at_ovs(port)

        self.assertEqual(port_iface, ret)
        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['get', 'Interface', port, 'ofport'])

    def test_get_ovs_ports_info(self):
        bridge = 'fake-bridge'
        bridge_ports = ['br-ex']
        self.mock_ovs_vsctl.ovs_vsctl.return_value = bridge_ports

        ret = ovs_utils.get_ovs_ports_info(bridge)

        self.assertEqual(bridge_ports, ret)
        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['list-ports', bridge])

    def test_get_ovs_patch_port_ofport(self):
        patch = 'fake-patch'
        ofport = ['1']
        self.mock_ovs_vsctl.ovs_vsctl.return_value = ofport

        ret = ovs_utils.get_ovs_patch_port_ofport(patch)

        self.assertEqual(ofport[0], ret)
        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['get', 'Interface', 'patch-fake-patch-to-br-int', 'ofport'])

    def test_get_ovs_patch_port_ofport_exception(self):
        patch = 'fake-patch'
        self.mock_ovs_vsctl.ovs_vsctl.side_effect = Exception

        self.assertRaises(agent_exc.PortNotFound,
                          ovs_utils.get_ovs_patch_port_ofport, patch)

        expected_calls = [
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport'])]
        self.mock_ovs_vsctl.ovs_vsctl.assert_has_calls(expected_calls)

    def test_get_ovs_patch_port_ofport_no_port(self):
        patch = 'fake-patch'
        ofport = ['[]']
        self.mock_ovs_vsctl.ovs_vsctl.return_value = ofport

        self.assertRaises(agent_exc.PortNotFound,
                          ovs_utils.get_ovs_patch_port_ofport, patch)

        expected_calls = [
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport']),
            mock.call(['get', 'Interface',
                       'patch-fake-patch-to-br-int', 'ofport'])]
        self.mock_ovs_vsctl.ovs_vsctl.assert_has_calls(expected_calls)

    @mock.patch.object(ovs_utils, 'get_bridge_flows')
    def test_remove_extra_ovs_flows(self, mock_flows):
        port_iface = '1'
        extra_port_iface = '2'
        extra_mac = 'ff:ee:dd:cc:bb:aa'
        self.flows_info[self.bridge]['in_port'] = {port_iface}
        self.flows_info[self.bridge]['mac'] = self.mac
        expected_flow = ("cookie={},priority=900,ip,in_port={} "
                         "actions=mod_dl_dst:{},NORMAL".format(
                             self.cookie, port_iface, self.mac))
        expected_flow_v6 = ("cookie={},priority=900,ipv6,in_port={} "
                            "actions=mod_dl_dst:{},NORMAL".format(
                                self.cookie, port_iface, self.mac))
        extra_flow = ("cookie={},priority=900,ip,in_port={} "
                      "actions=mod_dl_dst:{},NORMAL".format(
                          self.cookie, extra_port_iface, extra_mac))
        mock_flows.return_value = [expected_flow, expected_flow_v6, extra_flow]

        # Invoke the method
        ovs_utils.remove_extra_ovs_flows(self.flows_info, self.bridge,
                                         self.cookie)

        expected_del_flow = ('%s,ip,in_port=%s' % (self.cookie_id,
                                                   extra_port_iface))
        self.mock_ovs_vsctl.ovs_ofctl.assert_called_once_with(
            ['del-flows', self.bridge, expected_del_flow])
        mock_flows.assert_called_once_with(self.bridge, self.cookie_id)

    def test_ensure_flow(self):
        bridge = 'fake-bridge'
        flow = 'fake-flow'

        ovs_utils.ensure_flow(bridge, flow)

        self.mock_ovs_vsctl.ovs_ofctl.assert_called_once_with(
            ['add-flow', bridge, flow])

    @mock.patch.object(ovs_utils, 'get_device_port_at_ovs')
    @mock.patch.object(linux_net, 'get_interface_address')
    @mock.patch.object(linux_net, 'get_ip_version')
    def _test_ensure_evpn_ovs_flow(self, mock_ip_version, mock_nic_address,
                                   mock_ofport, ip_version, strip_vlan=False):
        address = '00:00:00:00:00:00'
        mock_ip_version.return_value = ip_version
        mock_nic_address.return_value = address
        port = 'fake-port'
        port_dst = 'fake-port-dst'
        ovs_port = constants.OVS_PATCH_PROVNET_PORT_PREFIX + 'fake-port'
        port_iface = '1'
        ovs_port_iface = '2'
        net = 'fake-net'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [
            '%s\n%s\n' % (port, ovs_port)]
        self.mock_ovs_vsctl.ovs_ofctl.return_value = None
        mock_ofport.side_effect = (ovs_port_iface, port_iface)

        # Invoke the method
        ovs_utils.ensure_evpn_ovs_flow(
            self.bridge, self.cookie, self.mac, port, port_dst, net,
            strip_vlan=strip_vlan)

        mock_ip_version.assert_called_once_with(net)
        strip_vlan_opt = 'strip_vlan,' if strip_vlan else ''
        if ip_version == constants.IP_VERSION_4:
            expected_flow = (
                "cookie={},priority=1000,ip,in_port={},dl_src:{},nw_src={}"
                "actions=mod_dl_dst:{},{}output={}".format(
                    self.cookie, ovs_port_iface, self.mac, net, address,
                    strip_vlan_opt, port_iface))
        else:
            expected_flow = (
                "cookie={},priority=1000,ipv6,in_port={},dl_src:{},"
                "ipv6_src={} actions=mod_dl_dst:{},{}output={}".format(
                    self.cookie, ovs_port_iface, self.mac, net, address,
                    strip_vlan_opt, port_iface))
        vsctl_expected_calls = [
            mock.call(['list-ports', self.bridge])]
        ofctl_expected_calls = [
            mock.call(['add-flow', self.bridge, expected_flow])]
        self.mock_ovs_vsctl.ovs_vsctl.assert_has_calls(vsctl_expected_calls)
        self.mock_ovs_vsctl.ovs_ofctl.assert_has_calls(ofctl_expected_calls)
        expected_calls_ofport = [mock.call(ovs_port), mock.call(port)]
        mock_ofport.assert_has_calls(expected_calls_ofport)

    def test_ensure_evpn_ovs_flow_ipv4(self):
        self._test_ensure_evpn_ovs_flow(ip_version=constants.IP_VERSION_4)

    def test_ensure_evpn_ovs_flow_ipv4_strip_vlan(self):
        self._test_ensure_evpn_ovs_flow(
            ip_version=constants.IP_VERSION_4, strip_vlan=True)

    def test_ensure_evpn_ovs_flow_ipv6(self):
        self._test_ensure_evpn_ovs_flow(ip_version=constants.IP_VERSION_6)

    def test_ensure_evpn_ovs_flow_ipv6_strip_vlan(self):
        self._test_ensure_evpn_ovs_flow(
            ip_version=constants.IP_VERSION_6, strip_vlan=True)

    def test_ensure_evpn_ovs_flow_no_ovs_ports(self):
        port = 'non-patch-provnet-port'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [port]

        ret = ovs_utils.ensure_evpn_ovs_flow(
            self.bridge, self.cookie, self.mac, port, 'fake-port-dst',
            'fake-net')

        self.assertIsNone(ret)
        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['list-ports', self.bridge])

    @mock.patch.object(ovs_utils, 'get_device_port_at_ovs')
    def test_remove_evpn_router_ovs_flows(self, mock_ofport):
        ovs_port = constants.OVS_PATCH_PROVNET_PORT_PREFIX + 'fake-port'
        ovs_port_iface = '1'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [ovs_port]
        self.mock_ovs_vsctl.ovs_ofctl.side_effect = (None, None)
        mock_ofport.return_value = ovs_port_iface

        # Invoke the method
        ovs_utils.remove_evpn_router_ovs_flows(
            self.bridge, self.cookie, self.mac)

        expected_flow = '{},ip,in_port={},dl_src:{}'.format(
            self.cookie_id, ovs_port_iface, self.mac)
        expected_flow_v6 = '{},ipv6,in_port={},dl_src:{}'.format(
            self.cookie_id, ovs_port_iface, self.mac)

        vsctl_expected_calls = [
            mock.call(['list-ports', self.bridge])]
        ofctl_expected_calls = [
            mock.call(['del-flows', self.bridge, expected_flow]),
            mock.call(['del-flows', self.bridge, expected_flow_v6])]

        self.mock_ovs_vsctl.ovs_vsctl.assert_has_calls(vsctl_expected_calls)
        self.mock_ovs_vsctl.ovs_ofctl.assert_has_calls(ofctl_expected_calls)

        mock_ofport.assert_called_once_with(ovs_port)

    def test_remove_evpn_router_ovs_flows_no_ovs_port(self):
        port = 'non-patch-provnet-port'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [port]

        ret = ovs_utils.remove_evpn_router_ovs_flows(
            self.bridge, self.cookie, self.mac)

        self.assertIsNone(ret)
        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['list-ports', self.bridge])

    @mock.patch.object(ovs_utils, 'get_device_port_at_ovs')
    @mock.patch.object(linux_net, 'get_ip_version')
    def _test_remove_evpn_network_ovs_flow(self, mock_ip_version, mock_ofport,
                                           ip_version):
        ovs_port = constants.OVS_PATCH_PROVNET_PORT_PREFIX + 'fake-port'
        ovs_port_iface = '1'
        net = 'fake-net'
        mock_ip_version.return_value = ip_version
        mock_ofport.return_value = ovs_port_iface
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [ovs_port]
        self.mock_ovs_vsctl.ovs_ofctl.side_effect = None

        ovs_utils.remove_evpn_network_ovs_flow(
            self.bridge, self.cookie, self.mac, net)

        if ip_version == constants.IP_VERSION_6:
            expected_flow = ("{},ipv6,in_port={},dl_src:{},ipv6_src={}".format(
                             self.cookie_id, ovs_port_iface, self.mac, net))
        else:
            expected_flow = ("{},ip,in_port={},dl_src:{},nw_src={}".format(
                             self.cookie_id, ovs_port_iface, self.mac, net))

        vsctl_expected_calls = [
            mock.call(['list-ports', self.bridge])]
        ofctl_expected_calls = [
            mock.call(['del-flows', self.bridge, expected_flow])]
        self.mock_ovs_vsctl.ovs_vsctl.assert_has_calls(vsctl_expected_calls)
        self.mock_ovs_vsctl.ovs_ofctl.assert_has_calls(ofctl_expected_calls)
        mock_ip_version.assert_called_once_with(net)

    def test_remove_evpn_network_ovs_flow_ipv4(self):
        self._test_remove_evpn_network_ovs_flow(
            ip_version=constants.IP_VERSION_4)

    def test_remove_evpn_network_ovs_flow_ipv6(self):
        self._test_remove_evpn_network_ovs_flow(
            ip_version=constants.IP_VERSION_6)

    def test_remove_evpn_network_ovs_flow_no_ovs_port(self):
        port = 'non-patch-provnet-port'
        self.mock_ovs_vsctl.ovs_vsctl.return_value = [port]

        ovs_utils.remove_evpn_network_ovs_flow(
            self.bridge, self.cookie, self.mac, 'fake-net')

        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(
            ['list-ports', self.bridge])

    def _test_add_device_to_ovs_bridge(self, vlan_tag=False):
        device = 'ethX'
        vtag = '1001' if vlan_tag else None

        ovs_utils.add_device_to_ovs_bridge(device, self.bridge, vlan_tag=vtag)

        expected_args = ['--may-exist', 'add-port', self.bridge, device]
        if vlan_tag:
            expected_args.append('tag=%s' % vtag)

        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(expected_args)

    def test_add_device_to_ovs_bridge(self):
        self._test_add_device_to_ovs_bridge()

    def test_add_device_to_ovs_bridge_vlan_tag(self):
        self._test_add_device_to_ovs_bridge(vlan_tag=True)

    def _test_del_device_from_ovs_bridge(self, bridge=False):
        device = 'ethX'
        br = self.bridge if bridge else None

        ovs_utils.del_device_from_ovs_bridge(device, bridge=br)

        expected_args = ['--if-exists', 'del-port']
        if bridge:
            expected_args.append(br)
        expected_args.append(device)

        self.mock_ovs_vsctl.ovs_vsctl.assert_called_once_with(expected_args)

    def test_del_device_from_ovs_bridge(self):
        self._test_del_device_from_ovs_bridge()

    def test_del_device_from_ovs_bridge_specifying_bridge(self):
        self._test_del_device_from_ovs_bridge(bridge=True)

    def test_del_flow(self):
        flow = ('cookie=0x3e6, duration=11.647s, table=0, n_packets=0, '
                'n_bytes=0, idle_age=3378, priority=1000,ip,dl_src=fa:16:3e'
                ':15:9e:f0,nw_src=20.0.0.0/24 actions=mod_dl_dst:d2:33:c5:'
                'fd:7c:42,output:3,in_port=1')
        ovs_utils.del_flow(flow, self.bridge, self.cookie)

        expected_flow = ('{},priority=1000,ip,dl_src=fa:16:3e:15:9e:f0,'
                         'nw_src=20.0.0.0/24'.format(self.cookie_id))
        self.mock_ovs_vsctl.ovs_ofctl.assert_called_once_with(
            ['--strict', 'del-flows', self.bridge, expected_flow])

    def test_get_flow_info(self):
        flow = ('cookie=0x3e6, duration=11.647s, table=0, n_packets=0, '
                'n_bytes=0, idle_age=3378, priority=1000,ip,dl_src=fa:16:3e'
                ':15:9e:f0,nw_src=20.0.0.0/24 actions=mod_dl_dst:d2:33:c5:'
                'fd:7c:42,output:3,in_port=1')

        ret = ovs_utils.get_flow_info(flow)

        expected_ret = {'ipv6_src': None, 'mac': 'fa:16:3e:15:9e:f0',
                        'nw_src': '20.0.0.0/24', 'port': '3'}
        self.assertEqual(expected_ret, ret)

    def test_get_flow_info_ipv6(self):
        flow = ('cookie=0x3e6, duration=9.275s, table=0, n_packets=0, '
                'n_bytes=0, idle_age=14326, priority=1000,ipv6,in_port=1,'
                'dl_src=fa:16:3e:15:9e:f0,ipv6_src=fdaa:4ad8:e8fb::/64 '
                'actions=mod_dl_dst:d2:33:c5:fd:7c:42,output:3')

        ret = ovs_utils.get_flow_info(flow)

        expected_ret = {'ipv6_src': 'fdaa:4ad8:e8fb::/64',
                        'mac': 'fa:16:3e:15:9e:f0', 'nw_src': None,
                        'port': '3'}
        self.assertEqual(expected_ret, ret)


class TestOvsIdl(test_base.TestCase):

    def setUp(self):
        super(TestOvsIdl, self).setUp()
        self.ovs_idl = ovs_utils.OvsIdl()
        self.ovs_idl.idl_ovs = mock.Mock()
        self.execute_ref = self.ovs_idl.idl_ovs.db_get.return_value.execute

    @mock.patch('ovsdbapp.backend.ovs_idl.connection.Connection')
    @mock.patch('ovs.db.idl.Idl')
    @mock.patch('ovsdbapp.backend.ovs_idl.idlutils.get_schema_helper')
    def test_start(self, mock_schema_helper, mock_idl, mock_conn):
        conn_str = 'fake-connection'
        self.ovs_idl.start(conn_str)

        mock_schema_helper.assert_called_once_with(conn_str, 'Open_vSwitch')
        helper = mock_schema_helper.return_value
        expected_calls = [
            mock.call('Open_vSwitch'), mock.call('Bridge'),
            mock.call('Port'), mock.call('Interface')]
        helper.register_table.assert_has_calls(expected_calls)
        mock_idl.assert_called_once_with(conn_str, helper)
        mock_conn.assert_called_once_with(
            mock_idl.return_value, timeout=mock.ANY)
        # Assert the OvsdbIdl instance was created
        self.assertIsInstance(self.ovs_idl.idl_ovs, idl_ovs.OvsdbIdl)

    def _test_ovs_ext_ids_getters(self, method, row, expected_return):
        self.execute_ref.return_value = row
        ret = method()
        self.assertEqual(expected_return, ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')

    def test_get_own_chassis_id(self):
        expected_return = 'fake-sys'
        row = {'system-id': expected_return}
        self._test_ovs_ext_ids_getters(
            self.ovs_idl.get_own_chassis_id, row, expected_return)

    def test_get_own_chassis_name(self):
        expected_return = 'fake-name'
        row = {'hostname': expected_return}
        self._test_ovs_ext_ids_getters(
            self.ovs_idl.get_own_chassis_name, row, expected_return)

    def test_get_ovn_remote(self):
        expected_return = 'fake-ovn-remote'
        row = {'ovn-remote': expected_return}
        self._test_ovs_ext_ids_getters(
            self.ovs_idl.get_ovn_remote, row, expected_return)

    def test_get_ovn_remote_nb(self):
        expected_return = 'fake-ovn-remote'
        row = {'ovn-nb-remote': expected_return}
        self.execute_ref.return_value = row
        ret = self.ovs_idl.get_ovn_remote(nb=True)
        self.assertEqual(expected_return, ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')

    def test_get_ovn_bridge_mappings(self):
        self.execute_ref.return_value = {
            'ovn-bridge-mappings': 'net0:bridge0,net1:bridge1, net2:bridge2'}
        ret = self.ovs_idl.get_ovn_bridge_mappings()
        self.assertEqual(['net0:bridge0', 'net1:bridge1', 'net2:bridge2'], ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')

    def test_get_ovn_bridge_mappings_not_set(self):
        self.execute_ref.return_value = {}
        ret = self.ovs_idl.get_ovn_bridge_mappings()
        self.assertEqual([], ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')

    def test_get_ovn_bridge_mappings_bridge(self):
        bridge = 'bgp1'
        self.execute_ref.return_value = {
            'ovn-bridge-mappings-bgp1':
            'net0:bridge0,net1:bridge1, net2:bridge2'}
        ret = self.ovs_idl.get_ovn_bridge_mappings(bridge=bridge)
        self.assertEqual(['net0:bridge0', 'net1:bridge1', 'net2:bridge2'], ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')

    def test_get_ovn_bridge_mappings_default_bridge(self):
        # bgp_chassis_id defaults to 'bgp'
        bridge = CONF.local_ovn_cluster.bgp_chassis_id
        self.execute_ref.return_value = {
            'ovn-bridge-mappings-bgp':
            'net0:bridge0,net1:bridge1, net2:bridge2'}
        ret = self.ovs_idl.get_ovn_bridge_mappings(bridge=bridge)
        self.assertEqual(['net0:bridge0', 'net1:bridge1', 'net2:bridge2'], ret)
        self.ovs_idl.idl_ovs.db_get.assert_called_once_with(
            'Open_vSwitch', '.', 'external_ids')



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
from ovsdbapp.backend.ovs_idl import event as row_event

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers.openstack.utils import driver_utils
from ovn_bgp_agent.drivers.openstack.utils import nat as nat_utils


LOG = logging.getLogger(__name__)


class Event(row_event.RowEvent):
    def __init__(self, agent, events, table, condition=None):
        self.agent = agent
        super().__init__(events, table, condition)

    def run(self, *args, **kwargs):
        try:
            self._run(*args, **kwargs)
        except Exception:
            LOG.exception("Unexpected exception while running the event "
                          "action")


class PortBindingChassisEvent(Event):
    def __init__(self, bgp_agent, events):
        table = 'Port_Binding'
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__


class OVNLBEvent(Event):
    def __init__(self, bgp_agent, events):
        table = 'Load_Balancer'
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__


class LogicalSwitchChassisEvent(Event):
    def __init__(self, bgp_agent, events):
        table = 'Logical_Switch'
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__


class LSPChassisEvent(Event):
    def __init__(self, bgp_agent, events):
        table = 'Logical_Switch_Port'
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__

    def _get_chassis(self, row, default_type=constants.OVN_VM_VIF_PORT_TYPE):
        return driver_utils.get_port_chassis(row, self.agent.chassis,
                                             default_port_type=default_type)


class LRPChassisEvent(Event):
    def __init__(self, bgp_agent, events):
        table = 'Logical_Router_Port'
        super().__init__(bgp_agent, events, table)
        self.event_name = self.__class__.__name__


class ChassisCreateEventBase(Event):
    table = None

    def __init__(self, bgp_agent):
        self.first_time = True
        events = (self.ROW_CREATE,)
        super().__init__(
            bgp_agent, events, self.table,
            (('name', '=', bgp_agent.chassis),))
        self.event_name = self.__class__.__name__

    def _run(self, event, row, old):
        if self.first_time:
            self.first_time = False
        else:
            LOG.info("Connection to OVSDB established, doing a full sync")
            self.agent.sync()


class ChassisCreateEvent(ChassisCreateEventBase):
    table = 'Chassis'


class ChassisPrivateCreateEvent(ChassisCreateEventBase):
    table = 'Chassis_Private'


class DnatSnatBaseEvent(Event):
    events = None

    def __init__(self, bgp_agent):
        table = 'NAT'
        super().__init__(
            bgp_agent,
            self.__class__.events,
            table,
            (('type', '=', 'dnat_and_snat'),))


class FipOnCRLRPBaseEvent(DnatSnatBaseEvent):
    """Base class for NAT event.

    It matches if the associated LSP is either virtual port or a normal port
    and if the gateway chassis port is on this chassis.
    """
    def match_fn(self, event, row, old):
        try:
            lsp_uuid = row.logical_port[0]
        except IndexError:
            LOG.warning("NAT entry %s has no associated logical port",
                        row.uuid)
            return False
        lsp = self.agent.nb_idl.lsp_get(
            lsp_uuid).execute(check_error=True)
        if not lsp:
            LOG.warning("Switch Port %s cannot be found" % lsp_uuid)
            return False

        if lsp.type not in [constants.OVN_VM_VIF_PORT_TYPE,
                            constants.OVN_VIRTUAL_VIF_PORT_TYPE]:
            return False

        if constants.OVN_FIP_NET_EXT_ID_KEY not in row.external_ids:
            LOG.error("NAT entry %(nat)s does not have %(fip_net)s set in "
                      "external_ids", {
                          'nat': row.uuid,
                          'fip_net': constants.OVN_FIP_NET_EXT_ID_KEY})
            return False

        gw_chassis = nat_utils.get_chassis_hosting_crlrp(row)

        return self.agent.chassis_id == gw_chassis



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import ipaddress
import threading

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from ovn_bgp_agent import constants
from ovn_bgp_agent.drivers import driver_api
from ovn_bgp_agent.drivers.openstack.utils import frr
from ovn_bgp_agent.drivers.openstack.utils import ovn
from ovn_bgp_agent.drivers.openstack.utils import ovs
from ovn_bgp_agent.drivers.openstack.watchers import base_watcher
from ovn_bgp_agent.drivers.openstack.watchers import evpn_watcher as \
    watcher
from ovn_bgp_agent.utils import helpers
from ovn_bgp_agent.utils import linux_net


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
# LOG.setLevel(logging.DEBUG)
# logging.basicConfig(level=logging.DEBUG)

OVN_TABLES = ["Port_Binding", "Chassis", "Datapath_Binding", "Chassis_Private"]
EVPN_INFO = collections.namedtuple(
    'EVPNInfo', ['vrf_name', 'lo_name', 'bridge_name', 'vxlan_name',
                 'veth_vrf', 'veth_ovs', 'vlan_name'])


class OVNEVPNDriver(driver_api.AgentDriverBase):

    def __init__(self):
        self.ovn_bridge_mappings = {}  # {'public': 'br-ex'}
        self.ovn_local_cr_lrps = {}
        self.ovn_local_lrps = {}
        # {'br-ex': [route1, route2]}
        self._ovn_routing_tables_routes = collections.defaultdict()
        self._ovn_exposed_evpn_ips = collections.defaultdict()

        self._sb_idl = None
        self._post_fork_event = threading.Event()

    @property
    def sb_idl(self):
        if not self._sb_idl:
            self._post_fork_event.wait()
        return self._sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self._sb_idl = val

    def start(self):
        self.ovs_idl = ovs.OvsIdl()
        self.ovs_idl.start(CONF.ovsdb_connection)
        self.chassis = self.ovs_idl.get_own_chassis_id()
        self.ovn_remote = self.ovs_idl.get_ovn_remote()
        LOG.debug("Loaded chassis %s.", self.chassis)

        self._post_fork_event.clear()

        events = self._get_events()
        self.sb_idl = ovn.OvnSbIdl(
            self.ovn_remote,
            chassis=self.chassis,
            tables=OVN_TABLES,
            events=events).start()

        # Now IDL connections can be safely used
        self._post_fork_event.set()

    def _get_events(self):
        return {watcher.PortBindingChassisCreatedEvent(self),
                watcher.PortBindingChassisDeletedEvent(self),
                watcher.SubnetRouterAttachedEvent(self),
                watcher.SubnetRouterDetachedEvent(self),
                watcher.TenantPortCreatedEvent(self),
                watcher.TenantPortDeletedEvent(self),
                base_watcher.ChassisCreateEvent(self),
                base_watcher.ChassisPrivateCreateEvent(self),
                watcher.LocalnetCreateDeleteEvent(self)}

    @lockutils.synchronized('evpn')
    def frr_sync(self):
        # Note(ltomasbo): There is no need for resync on this as there is
        # no base configuration to be made, but one added when subnets are
        # exposed, so the sync action takes care of it
        pass

    @lockutils.synchronized('evpn')
    def sync(self):
        self.ovn_local_cr_lrps = {}
        self.ovn_local_lrps = {}
        self._ovn_routing_tables_routes = collections.defaultdict()
        self._ovn_exposed_evpn_ips = collections.defaultdict()

        # 1) Get bridge mappings: xxxx:br-ex,yyyy:br-ex2
        bridge_mappings = self.ovs_idl.get_ovn_bridge_mappings()
        # 2) Get macs for bridge mappings
        for bridge_index, bridge_mapping in enumerate(bridge_mappings, 1):
            network, bridge = helpers.parse_bridge_mapping(bridge_mapping)
            if not network:
                continue
            self.ovn_bridge_mappings[network] = bridge

            linux_net.ensure_arp_ndp_enabled_for_bridge(bridge, bridge_index)

        # TO DO
        # add missing routes/ips for fips/provider VMs
        ports = self.sb_idl.get_ports_on_chassis(self.chassis)
        for port in ports:
            if port.type != constants.OVN_CHASSISREDIRECT_VIF_PORT_TYPE:
                continue
            self._expose_ip(port, cr_lrp=True)

        self._remove_extra_exposed_ips()
        self._remove_extra_routes()
        self._remove_extra_ovs_flows()
        self._remove_extra_vrfs()

    def _ensure_network_exposed(self, router_port, gateway):
        evpn_info = self.sb_idl.get_evpn_info_from_port_name(
            router_port.logical_port)
        if not evpn_info:
            LOG.debug("No EVPN information for LRP Port %s. "
                      "Not exposing it.", router_port)
            return

        gateway_ips = [ip.split('/')[0] for ip in gateway['ips']]
        try:
            router_port_ip = router_port.mac[0].strip().split(' ')[1]
        except IndexError:
            return
        router_ip = router_port_ip.split('/')[0]
        if router_ip in gateway_ips:
            return
        self.ovn_local_lrps[router_port.logical_port] = {
            'datapath': router_port.datapath,
            'ip': router_port_ip
            }
        datapath_bridge, vlan_tag = self._get_bridge_for_datapath(
            gateway['provider_datapath'])

        network_datapath = self.sb_idl.get_port_datapath(
            router_port.options['peer'])

        self._expose_subnet(router_port_ip, gateway_ips, gateway,
                            datapath_bridge, vlan_tag, network_datapath)

    def _get_bridge_for_datapath(self, datapath):
        network_name, network_tag = self.sb_idl.get_network_name_and_tag(
            datapath, self.ovn_bridge_mappings.keys())
        if network_name:
            if network_tag:
                return self.ovn_bridge_mappings[network_name], network_tag[0]
            return self.ovn_bridge_mappings[network_name], None
        return None, None

    @lockutils.synchronized('evpn')
    def expose_ip(self, row, cr_lrp=False):
        '''Advertice BGP route through EVPN.

        This methods ensures BGP advertises the IP through the required
        VRF/Tenant by using the specified VNI/VXLAN id.

        It relies on Zebra, which creates and advertises a route when an IP
        is added to a interface in the related VRF.
        '''
        self._expose_ip(row, cr_lrp)

    def _expose_ip(self, row, cr_lrp=False):
        if cr_lrp:
            cr_lrp_port_name = row.logical_port
            cr_lrp_port = row
        else:
            cr_lrp_port_name = 'cr-lrp-' + row.logical_port
            cr_lrp_port = self.sb_idl.get_port_if_local_chassis(
                cr_lrp_port_name, self.chassis)
            if not cr_lrp_port:
                # Not in local chassis, no need to proccess
                return

        _, cr_lrp_datapath = self.sb_idl.get_fip_associated(
            cr_lrp_port_name)
        if not cr_lrp_datapath:
            return

        if len(cr_lrp_port.mac[0].strip().split(' ')) < 2:
            return
        ips = cr_lrp_port.mac[0].strip().split(' ')[1:]

        if cr_lrp:
            evpn_info = self.sb_idl.get_evpn_info_from_port_name(
                cr_lrp_port_name)
        else:
            evpn_info = self.sb_idl.get_evpn_info(row)
        if not evpn_info:
            LOG.debug("No EVPN information for CR-LRP Port with IPs %s. "
                      "Not exposing it.", ips)
            return

        datapath_bridge, vlan_tag = self._get_bridge_for_datapath(
            cr_lrp_datapath)

        LOG.info("Adding BGP route for CR-LRP Port %s on AS %s and "
                 "VNI %s", ips, evpn_info['bgp_as'], evpn_info['vni'])
        evpn_devices = self._ensure_evpn_devices(datapath_bridge,
                                                 evpn_info['vni'],
                                                 vlan_tag)
        if not evpn_devices.vrf_name or not evpn_devices.lo_name:
            return

        self.ovn_local_cr_lrps[cr_lrp_port_name] = {
            'router_datapath': cr_lrp_port.datapath,
            'provider_datapath': cr_lrp_datapath,
            'ips': ips,
            'mac': cr_lrp_port.mac[0].strip().split(' ')[0],
            'vni': int(evpn_info['vni']),
            'bgp_as': evpn_info['bgp_as'],
            'lo': evpn_devices.lo_name,
            'bridge': evpn_devices.bridge_name,
            'vxlan': evpn_devices.vxlan_name,
            'vrf': evpn_devices.vrf_name,
            'veth_vrf': evpn_devices.veth_vrf,
            'veth_ovs': evpn_devices.veth_ovs,
            'vlan': evpn_devices.vlan_name
        }

        frr.vrf_reconfigure(evpn_info, action="add-vrf")

        self._connect_evpn_to_ovn(evpn_devices.vrf_name, evpn_devices.veth_vrf,
                                  evpn_devices.veth_ovs, ips, datapath_bridge,
                                  evpn_info['vni'], evpn_devices.vlan_name,
                                  vlan_tag)

        ips_without_mask = [ip.split("/")[0] for ip in ips]
        nei_dev = evpn_devices.vlan_name if vlan_tag else evpn_devices.veth_vrf
        for ip in ips_without_mask:
            linux_net.add_ip_nei(
                ip, self.ovn_local_cr_lrps[cr_lrp_port_name]['mac'], nei_dev)

        # Check if there are networks attached to the router,
        # and if so, add the needed routes/rules
        lrp_ports = self.sb_idl.get_lrp_ports_for_router(
            cr_lrp_port.datapath)
        for lrp in lrp_ports:
            if lrp.chassis or "chassis-redirect-port" in lrp.options.keys():
                continue
            self._ensure_network_exposed(
                lrp, self.ovn_local_cr_lrps[cr_lrp_port_name])

    @lockutils.synchronized('evpn')
    def withdraw_ip(self, row, cr_lrp=False):
        '''Withdraw BGP route through EVPN.

        This methods ensures BGP withdraw the IP advertised through the
        required VRF/Tenant by using the specified VNI/VXLAN id.

        It relies on Zebra, which cwithdraws the advertisement as son as the
        IP is deleted from the interface in the related VRF.
        '''
        if cr_lrp:
            cr_lrp_port_name = row.logical_port
        else:
            cr_lrp_port_name = 'cr-lrp-' + row.logical_port

        cr_lrp_info = self.ovn_local_cr_lrps.get(cr_lrp_port_name, {})
        if not cr_lrp_info:
            # This means it is in a different chassis
            return
        cr_lrp_datapath = cr_lrp_info.get('provider_datapath')
        if not cr_lrp_datapath:
            return

        ips = cr_lrp_info.get('ips')
        evpn_vni = cr_lrp_info.get('vni')
        if not evpn_vni:
            LOG.debug("No EVPN information for CR-LRP Port with IPs %s. "
                      "No need to withdraw it.", ips)
            return

        LOG.info("Delete BGP route for CR-LRP Port %s on VNI %s", ips,
                 evpn_vni)
        datapath_bridge, vlan_tag = self._get_bridge_for_datapath(
            cr_lrp_datapath)

        if vlan_tag:
            self._disconnect_evpn_from_ovn(evpn_vni, datapath_bridge, ips,
                                           vlan_tag=vlan_tag)
        else:
            cr_lrps_on_same_provider = [
                p for p in self.ovn_local_cr_lrps.values()
                if p['provider_datapath'] == cr_lrp_datapath]
            if (len(cr_lrps_on_same_provider) > 1):
                # NOTE: no need to remove the NDP proxy if there are other
                # cr-lrp ports on the same chassis connected to the same
                # provider flat network
                self._disconnect_evpn_from_ovn(evpn_vni, datapath_bridge, ips,
                                               cleanup_ndp_proxy=False)
            else:
                self._disconnect_evpn_from_ovn(evpn_vni, datapath_bridge, ips)

        nei_dev = cr_lrp_info['vlan'] if vlan_tag else cr_lrp_info['veth_vrf']
        for ip in ips:
            linux_net.del_ip_nei(ip, cr_lrp_info['mac'], nei_dev)

        self._remove_evpn_devices(evpn_vni)
        ovs.remove_evpn_router_ovs_flows(datapath_bridge,
                                         constants.OVS_VRF_RULE_COOKIE,
                                         cr_lrp_info.get('mac'))

        evpn_info = {'vni': evpn_vni, 'bgp_as': cr_lrp_info.get('bgp_as')}
        frr.vrf_reconfigure(evpn_info, action="del-vrf")

        try:
            del self.ovn_local_cr_lrps[cr_lrp_port_name]
        except KeyError:
            LOG.debug("Gateway port already cleanup from the agent: %s",
                      cr_lrp_port_name)

    @lockutils.synchronized('evpn')
    def expose_remote_ip(self, ips, row):
        if self.sb_idl.is_provider_network(row.datapath):
            return
        port_lrps = self.sb_idl.get_lrps_for_datapath(row.datapath)
        for port_lrp in port_lrps:
            if port_lrp in self.ovn_local_lrps.keys():
                evpn_info = self.sb_idl.get_evpn_info_from_port_name(port_lrp)
                if not evpn_info:
                    LOG.debug("No EVPN information for LRP Port %s. "
                              "Not exposing IPs: %s.", port_lrp, ips)
                    continue
                LOG.info("Add BGP route for tenant IP %s on chassis %s",
                         ips, self.chassis)
                lo_name = constants.OVN_EVPN_LO_PREFIX + str(evpn_info['vni'])
                linux_net.add_ips_to_dev(
                    lo_name, ips, clear_local_route_at_table=evpn_info['vni'])
                self._ovn_exposed_evpn_ips.setdefault(
                    lo_name, []).extend(ips)

    @lockutils.synchronized('evpn')
    def withdraw_remote_ip(self, ips, row):
        if self.sb_idl.is_provider_network(row.datapath):
            return
        port_lrps = self.sb_idl.get_lrps_for_datapath(row.datapath)
        for port_lrp in port_lrps:
            if port_lrp in self.ovn_local_lrps.keys():
                evpn_info = self.sb_idl.get_evpn_info_from_port_name(port_lrp)
                if not evpn_info:
                    LOG.debug("No EVPN information for LRP Port %s. "
                              "Not withdrawing IPs: %s.", port_lrp, ips)
                    continue
                LOG.info("Delete BGP route for tenant IP %s on chassis %s",
                         ips, self.chassis)
                lo_name = constants.OVN_EVPN_LO_PREFIX + str(evpn_info['vni'])
                linux_net.del_ips_from_dev(lo_name, ips)

    @lockutils.synchronized('evpn')
    def expose_subnet(self, row):
        evpn_info = self.sb_idl.get_evpn_info(row)
        ip = self.sb_idl.get_ip_from_port_peer(row)
        if not evpn_info:
            LOG.debug("No EVPN information for LRP Port %s. "
                      "Not exposing IPs: %s.", row.logical_port, ip)
            return

        lrp_logical_port = 'lrp-' + row.logical_port
        lrp_datapath = self.sb_idl.get_port_datapath(lrp_logical_port)

        cr_lrp = self.sb_idl.is_router_gateway_on_chassis(lrp_datapath,
                                                          self.chassis)
        if not cr_lrp:
            return

        LOG.info("Add IP Routes for network %s on chassis %s", ip,
                 self.chassis)
        self.ovn_local_lrps[lrp_logical_port] = {
            'datapath': lrp_datapath,
            'ip': ip
            }

        cr_lrp_info = self.ovn_local_cr_lrps.get(cr_lrp, {})
        cr_lrp_datapath = cr_lrp_info.get('provider_datapath')
        if not cr_lrp_datapath:
            LOG.info("Subnet not connected to the provider network. "
                     "No need to expose it through EVPN")
            return
        if (evpn_info['bgp_as'] != cr_lrp_info.get('bgp_as') or
                evpn_info['vni'] != cr_lrp_info.get('vni')):
            LOG.error("EVPN information at router port (vni: %s, as: %s) does"
                      " not match with information at subnet gateway port:"
                      " %s", cr_lrp_info.get('vni'),
                      cr_lrp_info.get('bgp_as'), evpn_info)
            return

        cr_lrp_ips = [ip_address.split('/')[0]
                      for ip_address in cr_lrp_info.get('ips', [])]
        datapath_bridge, vlan_tag = self._get_bridge_for_datapath(
            cr_lrp_datapath)

        self._expose_subnet(ip, cr_lrp_ips, cr_lrp_info, datapath_bridge,
                            vlan_tag, row.datapath)

    def _expose_subnet(self, router_interface, cr_lrp_ips, cr_lrp_info,
                       datapath_bridge, vlan_tag, network_datapath):
        router_interface_ip_version = linux_net.get_ip_version(
            router_interface)
        if vlan_tag:
            dev = cr_lrp_info['vlan']
            dev_ovs = dev
            strip_vlan = True
        else:
            dev = cr_lrp_info['veth_vrf']
            dev_ovs = cr_lrp_info['veth_ovs']
            strip_vlan = False

        for cr_lrp_ip in cr_lrp_ips:
            if (linux_net.get_ip_version(cr_lrp_ip) ==
                    router_interface_ip_version):
                linux_net.add_ip_route(
                    self._ovn_routing_tables_routes,
                    router_interface.split("/")[0],
                    cr_lrp_info['vni'],
                    dev,
                    mask=router_interface.split("/")[1],
                    via=cr_lrp_ip)
                break

        if router_interface_ip_version == constants.IP_VERSION_6:
            net_ip = '{}'.format(ipaddress.IPv6Network(
                router_interface, strict=False))
        else:
            net_ip = '{}'.format(ipaddress.IPv4Network(
                router_interface, strict=False))

        # NOTE(ltomasbo): strip_vlan is used for subnets/routers associated to
        # provider vlan networks assuming the EVPN VXLAN header is replacing
        # the vlan id in the fabric. If that is not the case, we could simply
        # set this to False in all the cases and have the traffic sent with
        # both vxlan header (for the EVPN) plus the vlan header (related to
        # the provider vlan id being used)
        ovs.ensure_evpn_ovs_flow(datapath_bridge,
                                 constants.OVS_VRF_RULE_COOKIE,
                                 cr_lrp_info['mac'],
                                 dev_ovs,
                                 dev,
                                 net_ip,
                                 strip_vlan=strip_vlan)

        # Check if there are VMs on the network
        # and if so expose the route
        if not network_datapath:
            return
        ports = self.sb_idl.get_ports_on_datapath(
            network_datapath)
        for port in ports:
            if (port.type not in (constants.OVN_VM_VIF_PORT_TYPE,
                                  constants.OVN_VIRTUAL_VIF_PORT_TYPE) or
                (port.type == constants.OVN_VM_VIF_PORT_TYPE and
                 not port.chassis)):
                continue
            try:
                port_ips = port.mac[0].strip().split(' ')[1:]
            except IndexError:
                continue

            for port_ip in port_ips:
                # Only adding the port ips that match the lrp
                # IP version
                port_ip_version = linux_net.get_ip_version(port_ip)
                if port_ip_version == router_interface_ip_version:
                    linux_net.add_ips_to_dev(
                        cr_lrp_info['lo'], [port_ip],
                        clear_local_route_at_table=cr_lrp_info['vni'])
                    self._ovn_exposed_evpn_ips.setdefault(
                        cr_lrp_info['lo'], []).extend([port_ip])

    @lockutils.synchronized('evpn')
    def withdraw_subnet(self, row):
        lrp_logical_port = 'lrp-' + row.logical_port
        lrp_datapath = self.ovn_local_lrps.get(lrp_logical_port, {}).get(
            'datapath')
        ip = self.ovn_local_lrps.get(lrp_logical_port, {}).get('ip')
        if not lrp_datapath:
            return

        cr_lrp = self.sb_idl.is_router_gateway_on_chassis(lrp_datapath,
                                                          self.chassis)
        if not cr_lrp:
            return

        LOG.info("Delete IP Routes for network %s on chassis %s", ip,
                 self.chassis)

        cr_lrp_info = self.ovn_local_cr_lrps.get(cr_lrp, {})
        cr_lrp_datapath = cr_lrp_info.get('provider_datapath')
        if not cr_lrp_datapath:
            LOG.info("Subnet not connected to the provider network. "
                     "No need to withdraw it from EVPN")
            return
        cr_lrp_ips = [ip_address.split('/')[0]
                      for ip_address in cr_lrp_info.get('ips', [])]
        datapath_bridge, vlan_tag = self._get_bridge_for_datapath(
            cr_lrp_datapath)

        if vlan_tag:
            dev = cr_lrp_info['vlan']
        else:
            dev = cr_lrp_info['veth_vrf']

        ip_version = linux_net.get_ip_version(ip)
        for cr_lrp_ip in cr_lrp_ips:
            if linux_net.get_ip_version(cr_lrp_ip) == ip_version:
                linux_net.del_ip_route(
                    self._ovn_routing_tables_routes,
                    ip.split("/")[0],
                    cr_lrp_info['vni'],
                    dev,
                    mask=ip.split("/")[1],
                    via=cr_lrp_ip)
                if (linux_net.get_ip_version(cr_lrp_ip) ==
                        constants.IP_VERSION_6):
                    net = ipaddress.IPv6Network(ip, strict=False)
                else:
                    net = ipaddress.IPv4Network(ip, strict=False)
                break

        ovs.remove_evpn_network_ovs_flow(datapath_bridge,
                                         constants.OVS_VRF_RULE_COOKIE,
                                         cr_lrp_info['mac'],
                                         '{}'.format(net))

        # Check if there are VMs on the network
        # and if so withdraw the routes
        vms_on_net = linux_net.get_exposed_ips_on_network(
            cr_lrp_info['lo'], net)
        linux_net.delete_exposed_ips(vms_on_net,
                                     cr_lrp_info['lo'])

        try:
            del self.ovn_local_lrps[lrp_logical_port]
        except KeyError:
            LOG.debug("Router Interface port already cleanup from the agent "
                      "%s", lrp_logical_port)

    def _ensure_evpn_devices(self, datapath_bridge, vni, vlan_tag):
        '''Create the needed devices for EVPN connectivity

        This method creates and associate the needed devices for EVPN
        connectivity. It creates:
        - VRF device
        - Linux Bridge device, associated to the VRF
        - VXLAN device, using loopback IP, associate to the bridge
        - Dummy device to expose the IPs, associated to the VRF
        - If vlan_tag, create vlan device on OVS bridge, associated to the VRF
        - If no vlan_tag, create veth pair, one end associated to the VRF

        param datapath_bridge: OVS bridge to connect the vlan device
        param vni: VNI number to use for vxlan tunnel ids and vrf routing table
        param vlan_tag: vlan id to use for connectivity

        return: a namedtuple with the name of the devices created: vrf_name,
        lo_name, bridge_name, vxlan_name, veth_vrf, veth_ovs, and vlan_name.
        '''
        # ensure vrf device.
        # NOTE: It uses vni id as table number
        vrf_name = constants.OVN_EVPN_VRF_PREFIX + str(vni)
        linux_net.ensure_vrf(vrf_name, vni)

        # ensure bridge device
        bridge_name = constants.OVN_EVPN_BRIDGE_PREFIX + str(vni)
        linux_net.ensure_bridge(bridge_name)
        # connect bridge to vrf
        linux_net.set_master_for_device(bridge_name, vrf_name)

        # ensure vxlan device
        vxlan_name = constants.OVN_EVPN_VXLAN_PREFIX + str(vni)

        local_ip = CONF.evpn_local_ip
        if not local_ip:
            local_nic = 'lo'
            prefixlen_filter = 32  # assuming IPv4
            if CONF.evpn_nic:
                local_nic = CONF.evpn_nic
                prefixlen_filter = False
            # NOTE(ltomasbo): assuming only 1 IP on the device with /32 prefix
            local_ip = linux_net.get_nic_ip(local_nic, prefixlen_filter)[0]

        if not local_ip:
            LOG.error("EVPN device must have an IP associated for the "
                      "VXLAN local ip")
            return None, None
        linux_net.ensure_vxlan(vxlan_name, vni, local_ip,
                               CONF.evpn_udp_dstport)
        # connect vxlan to bridge
        linux_net.set_master_for_device(vxlan_name, bridge_name)

        # ensure dummy lo interface
        lo_name = constants.OVN_EVPN_LO_PREFIX + str(vni)
        linux_net.ensure_dummy_device(lo_name)
        # connect dummy to vrf
        linux_net.set_master_for_device(lo_name, vrf_name)

        if vlan_tag:
            vlan_name = constants.OVN_EVPN_VLAN_PREFIX + str(vni)
            # add vlan port to OVS bridge
            ovs.add_vlan_port_to_ovs_bridge(datapath_bridge, vlan_name,
                                            vlan_tag)
            linux_net.set_device_status(vlan_name, constants.LINK_UP)
            # connect vlan to vrf
            linux_net.set_master_for_device(vlan_name, vrf_name)
            # ensure proxy NDP is enabled for ipv6 traffic
            linux_net.enable_proxy_ndp(vlan_name)

            return EVPN_INFO(vrf_name, lo_name, bridge_name, vxlan_name, None,
                             None, vlan_name)
        else:
            # ensure veth-pair interfaces
            veth_vrf = constants.OVN_EVPN_VETH_VRF_PREFIX + str(vni)
            veth_ovs = constants.OVN_EVPN_VETH_OVS_PREFIX + str(vni)
            linux_net.ensure_veth(veth_vrf, veth_ovs)
            # connect veth to vrf
            linux_net.set_master_for_device(veth_vrf, vrf_name)

            return EVPN_INFO(vrf_name, lo_name, bridge_name, vxlan_name,
                             veth_vrf, veth_ovs, None)

    def _remove_evpn_devices(self, vni):
        vrf_name = constants.OVN_EVPN_VRF_PREFIX + str(vni)
        bridge_name = constants.OVN_EVPN_BRIDGE_PREFIX + str(vni)
        vxlan_name = constants.OVN_EVPN_VXLAN_PREFIX + str(vni)
        lo_name = constants.OVN_EVPN_LO_PREFIX + str(vni)
        veth_name = constants.OVN_EVPN_VETH_VRF_PREFIX + str(vni)
        vlan_name = constants.OVN_EVPN_VLAN_PREFIX + str(vni)

        for device in [lo_name, vrf_name, bridge_name, vxlan_name, veth_name,
                       vlan_name]:
            linux_net.delete_device(device)

    def _connect_evpn_to_ovn(self, vrf, veth_vrf, veth_ovs, ips,
                             datapath_bridge, vni, vlan, vlan_tag):
        # NOTE(ltomasbo): vlan device is already attached to ovs bridge
        # when created
        if not vlan_tag:
            # add veth to ovs bridge
            ovs.add_device_to_ovs_bridge(veth_ovs, datapath_bridge)

        # add route for ip to ovs provider bridge (at the vrf routing table)
        for ip in ips:
            ip_without_mask = ip.split("/")[0]
            if vlan_tag:
                # ip route add GW_PORT_IP dev VLAN_DEVICE table VRF_TABLE_ID
                linux_net.add_ip_route(
                    self._ovn_routing_tables_routes, ip_without_mask,
                    vni, vlan)
                # add proxy ndp config for ipv6
                if (linux_net.get_ip_version(ip_without_mask) ==
                        constants.IP_VERSION_6):
                    linux_net.add_ndp_proxy(ip, vlan)
            else:
                linux_net.add_ip_route(
                    self._ovn_routing_tables_routes, ip_without_mask,
                    vni, veth_vrf)
                # add proxy ndp config for ipv6
                if (linux_net.get_ip_version(ip_without_mask) ==
                        constants.IP_VERSION_6):
                    linux_net.add_ndp_proxy(ip, datapath_bridge)

        # add unreachable route to vrf
        linux_net.add_unreachable_route(vrf)

    def _disconnect_evpn_from_ovn(self, vni, datapath_bridge, ips,
                                  vlan_tag=None, cleanup_ndp_proxy=True):
        if vlan_tag:
            # remove vlan from ovs bridge
            device = constants.OVN_EVPN_VLAN_PREFIX + str(vni)
        else:
            # remove veth from ovs bridge
            device = constants.OVN_EVPN_VETH_OVS_PREFIX + str(vni)
        ovs.del_device_from_ovs_bridge(device, datapath_bridge)

        linux_net.delete_routes_from_table(vni)

        if cleanup_ndp_proxy:
            for ip in ips:
                if linux_net.get_ip_version(ip) == constants.IP_VERSION_6:
                    linux_net.del_ndp_proxy(ip, datapath_bridge)

    def _remove_extra_vrfs(self):
        vrfs, los, bridges, vxlans, veths, vlans = ([], [], [], [], [], [])
        for cr_lrp_info in self.ovn_local_cr_lrps.values():
            vrfs.append(cr_lrp_info['vrf'])
            los.append(cr_lrp_info['lo'])
            bridges.append(cr_lrp_info['bridge'])
            vxlans.append(cr_lrp_info['vxlan'])
            veths.append(cr_lrp_info['veth_vrf'])
            vlans.append(cr_lrp_info['vlan'])

        filter_out = ["{}.{}".format(key, value[0]['vlan'])
                      for key, value in self._ovn_routing_tables_routes.items()
                      if value[0]['vlan']]

        interfaces = linux_net.get_interfaces(filter_out)
        for interface in interfaces:
            if (interface.startswith(constants.OVN_EVPN_VRF_PREFIX) and
                    interface not in vrfs):
                linux_net.delete_device(interface)
            elif (interface.startswith(constants.OVN_EVPN_LO_PREFIX) and
                    interface not in los):
                linux_net.delete_device(interface)
            elif (interface.startswith(constants.OVN_EVPN_BRIDGE_PREFIX) and
                    (interface not in bridges and
                     interface != constants.OVN_INTEGRATION_BRIDGE and
                     interface not in set(self.ovn_bridge_mappings.values()))):
                linux_net.delete_device(interface)
            elif (interface.startswith(constants.OVN_EVPN_VXLAN_PREFIX) and
                    interface not in vxlans):
                linux_net.delete_device(interface)
            elif (interface.startswith(constants.OVN_EVPN_VETH_VRF_PREFIX) and
                    interface not in veths):
                linux_net.delete_device(interface)
                ovs.del_device_from_ovs_bridge(interface)
            elif (interface.startswith(constants.OVN_EVPN_VLAN_PREFIX) and
                    interface not in vlans):
                ovs.del_device_from_ovs_bridge(interface)

    def _remove_extra_routes(self):
        table_ids = self._get_table_ids()
        vrf_routes = linux_net.get_routes_on_tables(table_ids)
        if not vrf_routes:
            return
        # remove from vrf_routes the routes that should be kept
        for device, routes_info in self._ovn_routing_tables_routes.items():
            for route_info in routes_info:
                oif = linux_net.get_interface_index(device)
                if 'gateway' in route_info['route'].keys():  # subnet route
                    possible_matchings = [
                        r for r in vrf_routes
                        if (r.get('dst') == route_info['route']['dst'] and
                            r['dst_len'] == route_info['route']['dst_len'] and
                            r.get('gateway') == (
                                route_info['route']['gateway']) and
                            r['table'] == route_info['route']['table'])]
                else:  # cr-lrp
                    possible_matchings = [
                        r for r in vrf_routes
                        if (r.get('dst') == route_info['route']['dst'] and
                            r['dst_len'] == route_info['route']['dst_len'] and
                            r.get('oif') == oif and
                            r['table'] == route_info['route']['table'])]
                for r in possible_matchings:
                    vrf_routes.remove(r)

        linux_net.delete_ip_routes(vrf_routes)

    def _remove_extra_ovs_flows(self):
        cr_lrp_mac_mappings = self._get_cr_lrp_mac_mapping()
        cookie_id = "cookie={}/-1".format(constants.OVS_VRF_RULE_COOKIE)
        for bridge in set(self.ovn_bridge_mappings.values()):
            current_flows = ovs.get_bridge_flows(bridge, filter_=cookie_id)
            for flow in current_flows:
                flow_info = ovs.get_flow_info(flow)
                if not flow_info.get('mac'):
                    ovs.del_flow(flow, bridge, constants.OVS_VRF_RULE_COOKIE)
                elif flow_info['mac'] not in cr_lrp_mac_mappings.keys():
                    ovs.del_flow(flow, bridge, constants.OVS_VRF_RULE_COOKIE)
                elif flow_info['port']:
                    if (not flow_info.get('nw_src') and not
                            flow_info.get('ipv6_src')):
                        ovs.del_flow(flow, bridge,
                                     constants.OVS_VRF_RULE_COOKIE)
                    else:
                        dev_info = cr_lrp_mac_mappings[flow_info['mac']]
                        if dev_info.get('vlan'):
                            dev = dev_info['vlan']
                            dev_ovs = dev
                        else:
                            dev = dev_info['veth_vrf']
                            dev_ovs = dev_info['veth_ovs']
                        dev_ovs_port = ovs.get_device_port_at_ovs(
                            dev_ovs)

                        if dev_ovs_port != flow_info['port']:
                            ovs.del_flow(flow, bridge,
                                         constants.OVS_VRF_RULE_COOKIE)
                        nw_src_ip = nw_src_mask = None
                        matching_dst = False
                        if flow_info.get('nw_src'):
                            nw_src_ip = flow_info['nw_src'].split('/')[0]
                            nw_src_mask = int(
                                flow_info['nw_src'].split('/')[1])
                        elif flow_info.get('ipv6_src'):
                            nw_src_ip = flow_info['ipv6_src'].split('/')[0]
                            nw_src_mask = int(
                                flow_info['ipv6_src'].split('/')[1])

                        for route_info in self._ovn_routing_tables_routes[
                                dev]:
                            if (route_info['route']['dst'] == nw_src_ip and
                                    route_info['route'][
                                        'dst_len'] == nw_src_mask):
                                matching_dst = True
                        if not matching_dst:
                            ovs.del_flow(flow, bridge,
                                         constants.OVS_VRF_RULE_COOKIE)

    def _remove_extra_exposed_ips(self):
        for lo, ips in self._ovn_exposed_evpn_ips.items():
            exposed_ips_on_device = linux_net.get_exposed_ips(lo)
            for ip in exposed_ips_on_device:
                if ip not in ips:
                    linux_net.del_ips_from_dev(lo, [ip])

    def _get_table_ids(self):
        table_ids = []
        for cr_lrp_info in self.ovn_local_cr_lrps.values():
            table_ids.append(cr_lrp_info['vni'])
        return table_ids

    def _get_cr_lrp_mac_mapping(self):
        mac_mappings = {}
        for cr_lrp_info in self.ovn_local_cr_lrps.values():
            mac_mappings[cr_lrp_info['mac']] = {
                'veth_vrf': cr_lrp_info['veth_vrf'],
                'veth_ovs': cr_lrp_info['veth_ovs'],
                'vlan': cr_lrp_info['vlan']}
        return mac_mappings



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from stevedore import driver as stevedore_driver


class AgentDriverBase(object, metaclass=abc.ABCMeta):
    """Base class for agent drivers.

    """

    @classmethod
    def get_instance(cls, specific_driver):
        agent_driver = stevedore_driver.DriverManager(
            namespace='ovn_bgp_agent.drivers',
            name=specific_driver,
            invoke_on_load=True
        ).driver

        return agent_driver

    @abc.abstractmethod
    def expose_ip(self, ip_address):
        raise NotImplementedError()

    @abc.abstractmethod
    def withdraw_ip(self, ip_address):
        raise NotImplementedError()

    @abc.abstractmethod
    def expose_remote_ip(self, ip_address):
        raise NotImplementedError()

    @abc.abstractmethod
    def withdraw_remote_ip(self, ip_address):
        raise NotImplementedError()

    @abc.abstractmethod
    def expose_subnet(self, subnet):
        raise NotImplementedError()

    @abc.abstractmethod
    def withdraw_subnet(self, subnet):
        raise NotImplementedError()
